#if !defined HATP_TASK_HH_
#define HATP_TASK_HH_

#include<string>
#include<vector>
#include<map>

#include "Entity.hh"

#define ACTION true
#define METHOD false
#define EMPTY_TASK "emptyTask"

class TaskList;
class LightEntity;
class Entity;
class LightWorldBase;
class WorldBase;
class WorldBaseReport;
class HTNBase;

/** Define the type for precondition function */
typedef bool (*precondFunc)(LightWorldBase const&, std::vector<LightEntity const*> const&);
/** Define the type for precondition function with a report */
typedef bool (*precondWithReportFunc)(LightWorldBase const&, std::vector<LightEntity const*> const&, WorldBaseReport&);
/** Define the type for empty function */
typedef bool (*emptyFunc)(LightWorldBase const&, std::vector<LightEntity const*> const&);
/** Define the type for decomposition function */
typedef std::vector<TaskList*> (*decompoFunc)(LightWorldBase const&, HTNBase const&, std::vector<LightEntity const*> const&);
/** Define the type for effect function */
typedef void (*effectsFunc)(LightWorldBase&, std::vector<LightEntity*> const&);
/** Define the type for effect function with a report */
typedef void (*effectsWithReportFunc)(LightWorldBase&, std::vector<LightEntity*> const&, WorldBaseReport&);
/** Define the type for cost function */
typedef float (*costFunc)(LightWorldBase const&, std::vector<LightEntity const*> const&);
/** Define the type for duration function */
typedef std::pair<double, double> (*timeFunc)(LightWorldBase const&, std::vector<LightEntity const*> const&);

/** Represent a task that can be either an action or a method (see HTN).
 */
class Task {

protected:
	/** Task name */
	std::string name;
	/** Number of arguments for the task's call */
	size_t nbArgs;
	/** Number of involver agents for the task's execution */
	size_t nbInvolvedAgents;
	/** The arguments' type, ordered by their order in the call */
	std::vector<std::string> argTypes;
	/** Tells if the task is an action or not. True if the task is an action. */
	bool isOperator;
	/** Pointer to the precondition function */
	precondFunc checkPreconds;
	/** Pointer to the precondition function (with report) */
	precondWithReportFunc checkPrecondsWithReport;
	/** Pointer to the emtpy function */
	emptyFunc checkEmpty;
	/** Pointer to the decomposition function */
	decompoFunc getDecompos;
	/** Pointer to the effect function */
	effectsFunc applyEffects;
	/** Pointer to the effect function (with report) */
	effectsWithReportFunc applyEffectsWithReport;
	/** Pointer to the cost function */
	costFunc computeCost;
	/** Pointer to the duration function */
	timeFunc computeTimeLimit;
	/** Whether it is a communcation task or not */
	bool communicationTask;

public:
	explicit Task();
	explicit Task(std::string const& sName, std::vector<std::string> const& svTypes, bool bOperator,
	precondFunc preconditionsFunction, precondWithReportFunc preconditionsWithReportFunction,
	emptyFunc emptyFunction, decompoFunc decompositionsFunction, effectsFunc effectsFunction,
	effectsWithReportFunc effectsWithReportFunction, costFunc costFunction, timeFunc timeFunction);
	Task(Task const& task_to_copy);
	bool checkTypes(std::vector<std::string> const& args, LightWorldBase const& wb) const;
	bool checkTypes(std::vector<std::string> const& args, WorldBase const& wb) const;
	bool checkTypes(std::vector<std::string> const& args, std::map<std::string, std::string> const& varAndTypes) const;
	std::string const& getName() const;
	size_t getNbArgs() const;
	std::vector<std::string> const& getArgTypes() const;
	bool isAction() const;
	decompoFunc const getDecompositionsFunction() const;
	precondFunc const getCheckingPreconditionsFunction() const;
	precondWithReportFunc const getCheckingPreconditionsWithReportFunction() const;
	emptyFunc const getCheckingEmptyFunction() const;
	effectsFunc const getApplyEffectsFunction() const;
	effectsWithReportFunc const getApplyEffectsWithReportFunction() const;
	costFunc const getComputeCostFunction() const;
	timeFunc const getComputeTimeFunction() const;
	std::string toString() const;
	size_t getNbInvolvedAgents() const;
	bool isCommunicationTask() const;
	void setCommunicationTask(bool f);
	~Task();
};

/** Default constructor.
 * Create an empty task.
 */
inline Task::Task() : name(EMPTY_TASK), nbArgs(0), nbInvolvedAgents(0), argTypes(std::vector<std::string>(0)), isOperator(ACTION),
	checkPreconds(NULL), checkPrecondsWithReport(NULL), checkEmpty(NULL), getDecompos(NULL), applyEffects(NULL), applyEffectsWithReport(NULL), 
	computeCost(NULL), computeTimeLimit(NULL), communicationTask(false) {}

/** Copy constructor.
 * @param task_to_copy Task to copy
 */
inline Task::Task(Task const& task_to_copy) : name(task_to_copy.getName()), nbArgs(task_to_copy.getNbArgs()),
	nbInvolvedAgents(task_to_copy.getNbInvolvedAgents()), argTypes(task_to_copy.getArgTypes()), isOperator(task_to_copy.isAction()), 
	checkPreconds(task_to_copy.getCheckingPreconditionsFunction()), checkPrecondsWithReport(task_to_copy.getCheckingPreconditionsWithReportFunction()),
	checkEmpty(task_to_copy.getCheckingEmptyFunction()), getDecompos(task_to_copy.getDecompositionsFunction()), 
	applyEffects(task_to_copy.getApplyEffectsFunction()), applyEffectsWithReport(task_to_copy.getApplyEffectsWithReportFunction()),
	computeCost(task_to_copy.getComputeCostFunction()), computeTimeLimit(task_to_copy.getComputeTimeFunction()),
	communicationTask(task_to_copy.isCommunicationTask()) {}

/** Get the name of the task
 * @return The task's name
 */
inline std::string const& Task::getName() const {
	return name;
}

/** Get the number of arguments for the task's call
 * @return Number of argument for the call
 */
inline size_t Task::getNbArgs() const {
	return nbArgs;
}

/** Get the arguments' type.
 * The set is ordered with the order of the arguments in the task call.
 * @return List of arguments' type
 */
inline std::vector<std::string> const& Task::getArgTypes() const {
	return argTypes;
}

/** Get the type of task : action or method,
 * @return True if the task is an action
 */
inline bool Task::isAction() const {
	return isOperator;
}

/** Get the pointer to the decomposition function.
 * @return Pointer to the function
 */
inline decompoFunc const Task::getDecompositionsFunction() const {
	return getDecompos;
}

/** Get the pointer to the precondition function.
 * @return Pointer to the function
 */
inline precondFunc const Task::getCheckingPreconditionsFunction() const {
	return checkPreconds;
}

/** Get the pointer to the precondition function (with report).
 * @return Pointer to the function
 */
inline precondWithReportFunc const Task::getCheckingPreconditionsWithReportFunction() const {
	return checkPrecondsWithReport;
}

/** Get the pointer to the empty function.
 * @return Pointer to the function
 */
inline emptyFunc const Task::getCheckingEmptyFunction() const {
	return checkEmpty;
}

/** Get the pointer to the effect function.
 * @return Pointer to the function
 */
inline effectsFunc const Task::getApplyEffectsFunction() const {
	return applyEffects;
}

/** Get the pointer to the effect function (with report).
 * @return Pointer to the function
 */
inline effectsWithReportFunc const Task::getApplyEffectsWithReportFunction() const {
	return applyEffectsWithReport;
}

/** Get the pointer to the cost function.
 * @return Pointer to the function
 */
inline costFunc const Task::getComputeCostFunction() const {
	return computeCost;
}

/** Get the pointer to the duration function.
 * @return Pointer to the function
 */
inline timeFunc const Task::getComputeTimeFunction() const {
	return computeTimeLimit;
}

/** Get the number of agents invovled in the execution of the task.
 * @return Number of invovled agents.
 */
inline size_t Task::getNbInvolvedAgents() const {
	return nbInvolvedAgents;
}

/** Tells whether this task is a communication task or not.
 * @return True if this is a communcation task
 */
inline bool Task::isCommunicationTask() const {
	return communicationTask;
}

/** Empty desctructor.
 */
inline Task::~Task() {}

#endif
