#if !defined HATP_WORLDBASE_HH_
#define HATP_WORLDBASE_HH_

#include<string>
#include<vector>
#include<map>

#include "Entity.hh"
#include "WorldBaseAgent.hh"

/** Represents the world as a set of Entity and a set of agents.
 * The agents are in fact WorldBaseAgent to only store their knowledge of the world.
 */
class WorldBase
{

private:
	/** Is used when returning a reference to an entity after an error in operator[](const char* c). */
	Entity defaultEntity;
	/** Contains the set of entities in the world. */
	std::map<std::string, Entity*> database;

	/** Tells whether the belief management is active or not */
	bool activeBM;
	/** The name of this database, is used in WorldBase::findForAgent(std::string agentName, std::string entityName) */
	std::string myselfName;
	/** Is the set of agents in the world and their representation of the world (See WorldBaseAgent). */
	std::map<std::string, WorldBaseAgent*> WBAgents;

public:
	explicit WorldBase();
	WorldBase(WorldBase const& wb);
	const std::map<std::string, Entity*>& getDatabase() const;
	void add(Entity const& e);
	bool containEntity(std::string const& sName) const;
	bool containEntity(std::string const& sName, std::string const& sType) const;
	Entity* find(std::string const& sName);
	Entity const* find(std::string const& sName) const;
	Entity* find(std::string const& sName, std::string const& sType);
	Entity const* find(std::string const& sName, std::string const& sType) const;
	std::vector<Entity*> findByType(std::string const& sType) const;
	std::vector<Entity*> findByAttribute(std::string const& attrName) const;

	const bool isActiveBM() const;
	const std::string getMyselfName() const;
	const std::map<std::string, WorldBaseAgent*>& getWBAgents() const;
	void setMyselfName(std::string agentName);
	void createWBForAgent(std::string const& agentName);
	void addForAgent(std::string const& agentName, Entity const& e);
	Entity* initForAgent(std::string agentName, std::string entityName);
	Entity* findForAgent(std::string agentName, std::string entityName);
	void updateForAgent(std::string a1, std::string a2, std::string e, std::string attr);

	std::string toString() const;
	Entity& operator[](const char* c);
	Entity& operator[](std::string const& s);
	Entity const& operator[](const char* c) const;
	Entity const& operator[](std::string const& s) const;
	WorldBase& operator=(WorldBase const& wb);
	void clearAllDynamicVectors();
	~WorldBase();
};

/** Default constructor.
 * All constructors set defaultEntity to an empty Entity.
 */
inline WorldBase::WorldBase() : defaultEntity(Entity()), activeBM(false) {}

/** Copy constructor.
 * All constructors set defaultEntity to an empty Entity.
 * @param wb World database to copy
 */
inline WorldBase::WorldBase(WorldBase const& wb) : defaultEntity(Entity()) {
	std::map<Entity*, Entity*> oldToNew; // to keep correspondence
	Entity *e;
	this->activeBM = wb.isActiveBM();
	this->myselfName = wb.getMyselfName();
	std::map<std::string, WorldBaseAgent*> const& atmp=wb.getWBAgents();
	for(std::map<std::string, WorldBaseAgent*>::const_iterator it = atmp.begin() ; it!=atmp.end() ; ++it) {
		this->WBAgents[it->first] = new WorldBaseAgent(*(it->second));
	}	
	std::map<std::string, Entity*> const& tmp=wb.getDatabase();
	for(std::map<std::string, Entity*>::const_iterator it = tmp.begin() ; it!=tmp.end() ; ++it) {
		e = new Entity(*(it->second));
		database[it->first] = e;
		oldToNew[it->second] = e;
	}
	// we need to update the Entity* attributes
	for(std::map<std::string, Entity*>::iterator itD = database.begin() ; itD!=database.end() ; ++itD)
		itD->second->updateAttrs(oldToNew);
}


/** Gives a reference to the entity database.
 * @return Entity database
 */
inline const std::map<std::string, Entity*>& WorldBase::getDatabase() const {
	return database;
}

/** Tells whether the Belief Management is active or not.
 * @return True if the Belief Management is active
 */
inline const bool WorldBase::isActiveBM() const {
	return activeBM;
}

/** Gives the name of this database.
 * @return The name of this database
 */
inline const std::string WorldBase::getMyselfName() const {
	return myselfName;
}

/** Gives the set of WorldBaseAgent.
 * The map key is the agents' name.
 * @return Agent Database
 */
inline const std::map<std::string, WorldBaseAgent*>& WorldBase::getWBAgents() const {
	return WBAgents;
}

/** Find an entity in a given database.
 * @param agentName Can either be the name of this database or the name of an agent from the world
 * @param entityName The entity to find
 * @return A pointer to the entity or NULL if the agent/entity does not exist
 */
inline Entity* WorldBase::findForAgent(std::string agentName, std::string entityName) {
	if(agentName == myselfName) {
		return this->find(entityName);
	}
	if(WBAgents.find(agentName) != WBAgents.end()) {
		return WBAgents.find(agentName)->second->find(entityName);
	}
	std::cerr << "ERROR when trying to access to an Agent DB for " << agentName << std::endl;
	return NULL;
}

/** Find an entity in the world database.
 * @param sName The entity to find
 * @return A pointer to the entity or NULL if the entity does not exist
 */
inline Entity* WorldBase::find(std::string const& sName) {
	std::map<std::string, Entity*>::iterator it = database.find(sName);
	if(it != database.end()) return it->second;
	std::cerr << "ERROR when trying to access to an unknown Entity \"" << sName;
	std::cerr << "\" in the WorldBase." << std::endl << "NULL returned instead" << std::endl;
	return NULL;
}

/** Find an entity in the world database.
 * @param sName The entity to find
 * @return A const pointer to the entity or NULL if the entity does not exist
 */
inline Entity const* WorldBase::find(std::string const& sName) const {
	std::map<std::string, Entity*>::const_iterator it = database.find(sName);
	if(it != database.end()) return it->second;
	std::cerr << "ERROR when trying to access to an unknown Entity \"" << sName;
	std::cerr << "\" in the WorldBase." << std::endl << "NULL returned instead" << std::endl;
	return NULL;
}

/** Assignement operator.
 * @param wb WorldBase to copy
 * @return Reference to this
 */
inline WorldBase& WorldBase::operator=(WorldBase const& wb) {
	// we empty database
	for(std::map<std::string, Entity*>::iterator itD = database.begin() ; itD != database.end() ; ++itD)
		delete itD->second;
	database.clear();
	std::map<Entity*, Entity*> oldToNew; // to keep correspondence
	Entity *e;
	this->myselfName = wb.getMyselfName();
	this->activeBM = wb.isActiveBM();
	std::map<std::string, WorldBaseAgent*> const& atmp=wb.getWBAgents();
	for(std::map<std::string, WorldBaseAgent*>::const_iterator it = atmp.begin() ; it!=atmp.end() ; ++it) {
		this->WBAgents[it->first] = new WorldBaseAgent(*(it->second));
	}	
	std::map<std::string, Entity*> tmp=wb.getDatabase();
	for(std::map<std::string, Entity*>::const_iterator it = tmp.begin() ; it!=tmp.end() ; ++it) {
		e = new Entity(*(it->second));
		database[it->first] = e;
		oldToNew[it->second] = e;
	}
	// we need to update the Entity* attributes
	for(std::map<std::string, Entity*>::iterator itD = database.begin() ; itD!=database.end() ; ++itD)
		itD->second->updateAttrs(oldToNew);
	return *this;
}

#endif
