/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_WORLDBASE_REPORT_HH_
#define HATP_WORLDBASE_REPORT_HH_

#include<string>
#include<vector>
#include<typeinfo>
#include<boost/regex.hpp>
#include<boost/lexical_cast.hpp>

#include "Entity.hh"

#define VECTOR_REMOVE "vector_remove"
#define VECTOR_ADD "vector_add"
#define VECTOR_TEST_IN "vector_test_in"
#define VECTOR_TEST_NOT_IN "vector_test_not_in"
#define ATOM_CHANGE "atom_change"
#define ATOM_TEST_EQUAL "atom_test_equal"
#define ATOM_TEST_DIFF "atom_test_diff"

static const std::string REPORT_space = "([[:space:]]+)";
static const std::string REPORT_term = "([[:word:]]+)";
static const std::string REPORT_emptyterm = "EMPTY";

// the syntax for value modification is: <entity> <attribute> value_change <old_value> <new_value>
static const std::string REPORT_atom_changeSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + ATOM_CHANGE + REPORT_space + REPORT_term + REPORT_space + REPORT_term;

// the syntax for adding value in a vector is: <entity> <attribute> vector_add <value>
static const std::string REPORT_vector_addSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + VECTOR_ADD + REPORT_space + REPORT_term;

// the syntax for removing value in a vector is: <entity> <attribute> vector_remove <value>
static const std::string REPORT_vector_removeSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + VECTOR_REMOVE + REPORT_space + REPORT_term;

// the syntax for tests on vectors is: <entity> <attribute> vector_test <value>
static const std::string REPORT_test_vector_inSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + VECTOR_TEST_IN + REPORT_space + REPORT_term;
static const std::string REPORT_test_vector_not_inSYNTAX = REPORT_term + REPORT_space + REPORT_term 
	+ REPORT_space + VECTOR_TEST_NOT_IN + REPORT_space + REPORT_term;

// the syntax for tests on values is: <entity> <attribute> value_test <value>
static const std::string REPORT_test_atom_equalSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + ATOM_TEST_EQUAL + REPORT_space + REPORT_term;
static const std::string REPORT_test_atom_diffSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + ATOM_TEST_DIFF + REPORT_space + REPORT_term;

static const boost::regex atom_changeREPORT(REPORT_atom_changeSYNTAX);
static const boost::regex vector_addREPORT(REPORT_vector_addSYNTAX);
static const boost::regex vector_removeREPORT(REPORT_vector_removeSYNTAX);
static const boost::regex atom_test_equalREPORT(REPORT_test_atom_equalSYNTAX);
static const boost::regex atom_test_diffREPORT(REPORT_test_atom_diffSYNTAX);
static const boost::regex vector_test_inREPORT(REPORT_test_vector_inSYNTAX);
static const boost::regex vector_test_not_inREPORT(REPORT_test_vector_not_inSYNTAX);

static const boost::regex attributeModifREPORT(REPORT_atom_changeSYNTAX + "|" + REPORT_vector_addSYNTAX + "|" + REPORT_vector_removeSYNTAX);
static const boost::regex attributeTestREPORT(REPORT_test_atom_equalSYNTAX +"|"+ REPORT_test_atom_diffSYNTAX +"|"+ REPORT_test_vector_inSYNTAX +"|"+ REPORT_test_vector_not_inSYNTAX);

std::vector<std::string> parseModificationReport(const std::string& report);
std::vector<std::string> parseTestReport(const std::string& report);

class WorldBaseReport
{
private:
	/**
	* the vector containing all test reports
	*/
	std::vector<std::string> tests;

	/**
	* the vector containing all modification reports
	*/
	std::vector<std::string> modifications;

public:
	WorldBaseReport();

	template<typename T> bool addAtomTestEqual(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addAtomTestEqual(std::string const& entity, const char* attribute, T val);
	template<typename T> bool addAtomTestDiff(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addAtomTestDiff(std::string const& entity, const char* attribute, T val);
	template<typename T> bool addVectorTestIn(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addVectorTestIn(std::string const& entity, const char* attribute, T val);
	template<typename T> bool addVectorTestNotIn(std::string const& entity, std::string const& attribute, T val);
	template<typename T> bool addVectorTestNotIn(std::string const& entity, const char* attribute, T val);
	template<typename T> void addAtomModifChange(std::string const& entity, std::string const& attribute, T oldVal, T newVal);
	template<typename T> void addAtomModifChange(std::string const& entity, const char* attribute, T oldVal, T newVal);
	template<typename T> void addVectorModifAdd(std::string const& entity, std::string const& attribute, T val);
	template<typename T> void addVectorModifAdd(std::string const& entity, const char* attribute, T val);
	template<typename T> void addVectorModifRemove(std::string const& entity, std::string const& attribute, T val);
	template<typename T> void addVectorModifRemove(std::string const& entity, const char* attribute, T val);
	std::vector<std::string> const& getTestsReport() const;
	std::vector<std::string> const& getModificationsReport() const;
	bool clearTests();
	bool clearModifications();
	bool clearAll();
	~WorldBaseReport();
};

/**
 * constructor: empty
 */
inline WorldBaseReport::WorldBaseReport()
{}

/**
 * add an atom test equal report in tests and return true
 */
template<typename T> bool WorldBaseReport::addAtomTestEqual(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += ATOM_TEST_EQUAL;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add an atom test equal report in tests and return true
 */
template<typename T> bool WorldBaseReport::addAtomTestEqual(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += ATOM_TEST_EQUAL;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add an atom test diff report in tests and return true
 */
template<typename T> bool WorldBaseReport::addAtomTestDiff(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += ATOM_TEST_DIFF;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add an atom test diff report in tests and return true
 */
template<typename T> bool WorldBaseReport::addAtomTestDiff(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += ATOM_TEST_DIFF;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add a vector test in report in tests and return true
 */
template<typename T> bool WorldBaseReport::addVectorTestIn(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_TEST_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add a vector test in report in tests and return true
 */
template<typename T> bool WorldBaseReport::addVectorTestIn(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_TEST_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add a vector test not in report in tests and return true
 */
template<typename T> bool WorldBaseReport::addVectorTestNotIn(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_TEST_NOT_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add a vector test not in report in tests and return true
 */
template<typename T> bool WorldBaseReport::addVectorTestNotIn(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an equal test , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_TEST_NOT_IN;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	tests.push_back(s);
	return true;
}

/**
 * add an atom modif change report in modifications
 */
template<typename T>
void WorldBaseReport::addAtomModifChange(std::string const& entity, std::string const& attribute, T oldVal, T newVal)
{
	std::string oldValue, newValue;
	try{
			 oldValue = boost::lexical_cast<std::string>(oldVal);
			 newValue = boost::lexical_cast<std::string>(newVal);
		 }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modification , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += ATOM_CHANGE;
	s += " ";
	if(oldValue.empty()) s += REPORT_emptyterm;
	else s += oldValue;
	s += " ";
	if(newValue.empty()) s += REPORT_emptyterm;
	else s += newValue;
	modifications.push_back(s);
}

/**
 * add an atom modif change report in modifications
 */
template<typename T> void WorldBaseReport::addAtomModifChange(std::string const& entity, const char* attribute, T oldVal, T newVal)
{
	std::string oldValue, newValue;

	try{
			 oldValue = boost::lexical_cast<std::string>(oldVal);
			 newValue = boost::lexical_cast<std::string>(newVal);
		 }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modification , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += ATOM_CHANGE;
	s += " ";
	if(oldValue.empty()) s += REPORT_emptyterm;
	else s += oldValue;
	s += " ";
	if(newValue.empty()) s += REPORT_emptyterm;
	else s += newValue;
	modifications.push_back(s);
}

/**
 * add a vector modif add report in modifications
 */
template<typename T> void WorldBaseReport::addVectorModifAdd(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an add in set modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_ADD;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}

/**
 * add a vector modif add report in modifications
 */
template<typename T> void WorldBaseReport::addVectorModifAdd(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an add in set modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_ADD;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}

/**
 * add a vector modif remove report in modifications
 */
template<typename T> void WorldBaseReport::addVectorModifRemove(std::string const& entity, std::string const& attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + attribute + " ";
	s += VECTOR_REMOVE;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}

/**
 * add a vector modif remove report in modifications
 */
template<typename T> void WorldBaseReport::addVectorModifRemove(std::string const& entity, const char* attribute, T val)
{
	std::string value;
	try{ value = boost::lexical_cast<std::string>(val); }
	catch(...) {
		std::cerr << "ERROR when making a report from an atom modifcation , type ";
		std::cerr << typeid(T).name() << " not convertible to a string" << std::endl;
	}
	std::string s = entity + " " + std::string(attribute) + " ";
	s += VECTOR_REMOVE;
	s += " ";
	if(value.empty()) s += REPORT_emptyterm;
	else s += value;
	modifications.push_back(s);
}


/**
 * return tests
 */
inline std::vector<std::string> const& WorldBaseReport::getTestsReport() const
{
	return tests;
}

/**
 * return modifications
 */
inline std::vector<std::string> const& WorldBaseReport::getModificationsReport() const
{
	return modifications;
}

/**
 * clear tests and return true
 */
inline bool WorldBaseReport::clearTests()
{
	tests.clear();
	return true;
}

/**
 * clear modifications and return true
 */
inline bool WorldBaseReport::clearModifications()
{
	modifications.clear();
	return true;
}

/**
 * clear tests and modifications and return true
 */
inline bool WorldBaseReport::clearAll()
{
	tests.clear();
	modifications.clear();
	return true;
}

/**
 * destructor: clear tests and modifications
 */
inline WorldBaseReport::~WorldBaseReport()
{
	tests.clear();
	modifications.clear();
}

#endif
